/* ttt-server.c - tic-tac-toe game server
 *
 * Copyright © 2005 Carl Worth
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Author: Carl Worth <cworth@cworth.org>
 */

#include "ttt.h"
#include "ttt-client.h"
#include "ttt-socket.h"

struct _ttt_server {
    pthread_mutex_t mutex;

    ttt_client_t **clients;
    int num_clients;
};

static void
ttt_server_init (ttt_server_t *server)
{
    pthread_mutex_init (&server->mutex, NULL);

    server->clients = NULL;
    server->num_clients = 0;
}

static void
ttt_server_add_client (ttt_server_t *server, ttt_client_t *client)
{
    pthread_mutex_lock (&server->mutex);

    server->num_clients++;
    server->clients = xrealloc (server->clients,
				server->num_clients * sizeof (ttt_client_t *));

    server->clients [server->num_clients - 1] = client;

    pthread_mutex_unlock (&server->mutex);
}

static void
_accept_client (void *closure, int client_socket)
{
    ttt_server_t *server = closure;
    ttt_client_t *client;
    
    client = ttt_client_create (server, client_socket);

    ttt_server_add_client (server, client);

    ttt_client_handle_requests (client);
}

static const char *WELCOME_MESSAGE = 
"Welcome to ttt-server. So far, this program is simply a demonstration\n"
"of a TCP/IP server capable of handling multiple simultaneous clients.\n"
"The server is currently listening on:\n"
"\n	%s:%s\n"
"\nTo test this program, simply connect a client to that host and port.\n"
"For example:\n"
"\n	telnet %s %s\n"
"\nOnce you have connected a client, the server will echo any characters\n"
"it receives back to the client as well as to stdout.\n"
"\nNote that to terminate the telnet client you type Control-], then\n"
"<Enter>, then \"close\" (and <Enter>) at the \"telnet> \" prompt.\n"
"\nHave fun!\n"
"-Carl\n"
"\nPS. The server does support multiple clients, but there is not yet any\n"
"interaction between the clients. The next step is probably to turn the\n"
"server into a simple chat server. After that, we should have the necessary\n"
"structure in place to start implementing the real tic-tac-toe protocol.\n\n";

int 
main (int argc, char **argv)
{
    ttt_args_t args;
    ttt_server_t server;
    int socket;

    ttt_args_parse (&args, argc, argv);

    if (args.log_file)
	xfreopen (args.log_file, "a", stderr);

    socket = ttt_socket_create_server (args.host, args.port);

    printf (WELCOME_MESSAGE, args.host, args.port, args.host, args.port);

    ttt_server_init (&server);

    while (1)
	ttt_socket_accept (socket, _accept_client, &server);

    /* We only reach here if something bad happened. */
    return 1;
}
